#ifndef __TASK_HPP__
#define __TASK_HPP__

#include <coroutine.hpp>
#include <result.hpp>
#include <promise.hpp>


namespace typon
{

  template <typename T = void>
  struct [[nodiscard]] task
  {
    struct promise_type;

    coroutine_handle<promise_type> _coroutine;
    result<T> _result;

    struct promise_type : promise<T>
    {
      coroutine_handle<> _continuation;

      task get_return_object() noexcept
      {
        task t {};
        auto h = coroutine_handle<promise_type>::from_promise(*this);
        t._coroutine = h;
        h.promise()._result = std::addressof(t._result);
        return t;
      }

      suspend_never initial_suspend() noexcept
      {
        return {};
      }

      auto final_suspend() noexcept
      {
        struct awaitable
        {
          bool _ready;

          bool await_ready() noexcept
          {
            return _ready;
          }

          coroutine_handle<> await_suspend(coroutine_handle<promise_type> coroutine) noexcept
          {
            return coroutine.promise()._continuation;
          }

          void await_resume() noexcept {}
        };

        return awaitable { !_continuation };
      }
    };

    auto operator co_await() && noexcept
    {
      struct awaitable
      {
        coroutine_handle<promise_type> _coroutine;
        result<T> * _result;
        bool _destroy;

        bool await_ready() noexcept
        {
          if (_result->ready())
          {
            return true;
          }
          _destroy = true;
          return false;
        }

        coroutine_handle<> await_suspend(coroutine_handle<> awaiting_coroutine) noexcept
        {
          _coroutine.promise()._continuation = awaiting_coroutine;
          return _coroutine;
        }

        decltype(auto) await_resume()
        {
          if (_destroy)
          {
            _coroutine.destroy();
          }
          return _result->get();
        }
      };

      return awaitable { _coroutine, std::addressof(_result), false };
    }

    decltype(auto) call() &&
    {
      return _result.get();
    }
  };
  
}

#endif // __TASK_HPP__
